#!/usr/bin/perl
#
# $Id: install.pl,v 1.3 2004/10/24 12:07:02 jvunder REL_0_3 $
# 
# HOSC install script
#
# Part of the Hybrid Oper Script Collection.
#

use strict;
use warnings;

use Getopt::Long;
use File::Copy;
use File::Basename;

use constant MODULES_DIR => 'HOSC';

my $opt;

my $res = GetOptions(
    'dir=s'       => \$opt->{'dir'},
    'force'       => \$opt->{'force'},
    'interactive' => \$opt->{'interactive'},
    'help'        => \$opt->{'help'},
);

if ($opt->{'help'}) {
    print usage();
    exit 0;
}

my $installdir = get_installdir();

if ($opt->{'interactive'}) {
    $opt->{'force'} = ask_force() unless defined $opt->{'force'};
}

print "Installing in $installdir\n";
print "Overwriting any existing files.\n" if $opt->{'force'};

if (!prepare_dir($installdir)) {
    print STDERR "Error preparing $installdir: $!. Bailing out.\n";
    exit 1;
}

if (!install_modules($installdir)) {
    print STDERR "Error installing modules.\n";
    exit 1;
}

if (!install_scripts($installdir)) {
    print STDERR "Error installing modules.\n";
    exit 1;
}

if (!install_rest($installdir)) {
    print STDERR "Error installing other files.\n";
    exit 1;
}

print "Installation successful.\n";

# ---------------------------------------------------------------------
# The end

exit 0;


# ---------------------------------------------------------------------
# Returns the dir to install. This was either given in a command line
# option, or it's asked of the user.

sub get_installdir {
    if (defined $opt->{'dir'}) {
        return $opt->{'dir'};
    }

    my $default_installdir = $ENV{'HOME'} . "/.irssi/scripts";

    if (defined $opt->{'interactive'}) {
        print "Install where? [$default_installdir] ";
        my $answer = <>;
        chomp $answer;
        if (length $answer) {
            return $answer;
        } else {
            return $default_installdir;
        }
    }

    return $default_installdir;
}

# ---------------------------------------------------------------------
# Ask the user whether they want to overwrite existing scripts; returns
# true if so and false otherwise.

sub ask_force {
    print "Do you want to overwrite existing script/module files (YN)? [N] ";
    my $answer = <>;
    chomp $answer;
    
    if (defined $answer && $answer =~ /^y/i) {
        return 1;
    }

    return 0;
}

# ---------------------------------------------------------------------
# Checks whether this dir already exists. If not, create it. Do the
# same for the 'HOSC' dir inside this dir.

sub prepare_dir {
    my ($dir) = @_;

    if (!-d $dir) {
        if (mkdir "$dir") {
            print "Created $dir\n";
        } else {
            print STDERR "Failed to create $dir: $!\n";
            return 0;
        }
    }
        
    if (-d "$dir/HOSC") {
        return 1;
    }

    if (mkdir "$dir/HOSC") {
        print "Created $dir/HOSC\n";
        return 1;
    } else {
        print STDERR "Failed to create $dir/HOSC: $!\n";
        return 0;
    }
    
}

# ---------------------------------------------------------------------

sub install_modules {
    my ($dir) = @_;

    return install_files(MODULES_DIR, $dir . "/" . MODULES_DIR, qr/\.pm$/);
}

# ---------------------------------------------------------------------

sub install_scripts {
    my ($dir) = @_;

    return install_files(".", $dir, qr/^ho_.*\.pl$/);
}

# ---------------------------------------------------------------------

sub install_rest {
    my ($dir) = @_;

    if (!install_files(".", $dir . "/" . MODULES_DIR, qr/^VERSION$/)) {
        return 0;
    }

    return 1;
}

# ---------------------------------------------------------------------
# Generic function to install all files in $source_dir matching
# $file_regexp into $target_dir.

sub install_files {
    my ($source_dir, $target_dir, $file_regexp) = @_;

    if (!-d $source_dir) {
        print STDERR "Source dir $source_dir does not exist.\n";
        return 0;
    }

    opendir D, $source_dir or die $!;
    my @files = readdir D;
    closedir D or die $!;

    for my $filename (@files) {
        next unless $filename =~ $file_regexp;
        copy_file($source_dir, $target_dir, $filename);
    }

    return 1;
}

# ---------------------------------------------------------------------
# Checks whether the way is free to copy the source to the target, and
# if so, does that.

sub copy_file {
    my ($source_dir, $installdir, $filename) = @_;

    # Find out what the version is of the existing file in the install dir.
    my $target_version = get_version("$installdir/$filename");

    # No version info found, or has the force option been used? 
    # Then just install the file.
    if (!defined $target_version || $opt->{'force'}) {
        print "$installdir/$filename\n";
        return do_copy("$source_dir/$filename", "$installdir/$filename");
    }

    my $source_version = get_version("$source_dir/$filename");

    # If target version is higher than the version we want to install,
    # don't overwrite it, unless the -f option was given.
    # If -i was given, ask the user.
    my $version_check = compare_versions($target_version, $source_version);
    if ($version_check == 1) {
        if ($opt->{'force'}) {
            print "$installdir/$filename\n";
            return do_copy("$source_dir/$filename", "$installdir/$filename");
        }
        
        if ($opt->{'interactive'}) {
            print "$installdir/$filename: Higher version ($target_version > ".
                "$source_version) already installed. Overwrite (YN) ? [N] ";
            my $answer = <>;
            if (length $answer && $answer =~ /^y/i) {
                print "$installdir/$filename\n";
                return do_copy("$source_dir/$filename", 
                               "$installdir/$filename");
            }

            return 1;
        } else {
            print "$installdir/$filename: higher version ($target_version > " .
                "$source_version) already installed. Not overwriting.\n";
            print "Use the -f option to force an overwrite.\n";
        }
    } elsif ($version_check == 0) {
        # Identical version, don't need to copy.
        #print "$filename: existing version $target_version identical to ".
        #    "new version. Not copying.\n";
        return 1;
    } else {
        # Lower version. Can overwrite.
        print "$installdir/$filename: Overwriting old version ".
            "$target_version with $source_version.\n";
        return do_copy("$source_dir/$filename", "$installdir/$filename");
    }
}

# ---------------------------------------------------------------------
# Actually does the file copy :)

sub do_copy {
    my ($source_file, $target_file) = @_;
    
    return unless defined $source_file && defined $target_file;

    if (copy($source_file, $target_file)) {
        return 1;
    } else {
        print STDERR "Error copying $source_file to $target_file: $!\n";
        return 0;
    }
}

# ---------------------------------------------------------------------
# Returns undef if there is no file $filename.
# Returns 0 if there is no line containing "$Revision: 1.3 $".
# Returns <Mv>.<mv> otherwise (Mv = Major version, mv = minor version).

sub get_version {
    my ($filename) = @_;

    return undef unless -f $filename;

    open F, $filename or return 0;

    while (my $line = <F>) {
        if ($line =~ /\$Revision: 1.3 $/) {
            close F;
            return $1;
        }
    }

    close F;
    return 0;
}

# ---------------------------------------------------------------------
# Returns 1  if $first > $second
# Returns 0  if $first == $second
# Returns -1 if $first < $second

sub compare_versions {
    my ($first, $second) = @_;

    # undef is considered to be version 0.
    $first  ||= "0.0";
    $second ||= "0.0";

    return 0 if "$first" eq "$second";

    my ($first_major,  $first_minor)  = $first  =~ /^(\d+)\.(\d+)$/;
    my ($second_major, $second_minor) = $second =~ /^(\d+)\.(\d+)$/;
    
    return  1 if $first_major > $second_major;
    return -1 if $first_major < $second_major;

    return  1 if $first_minor > $second_minor;
    return -1;
}

# ---------------------------------------------------------------------

sub usage {
    my $basename = basename($0);
    return <<EOU;
Usage: $basename -h
       $basename [-d <dir>] [-i] [-f]

    Installs the HOSC scripts and modules.

Options:
    -h  Print this help message and exit.
    -i  Use interactive mode.
    -f  Force overwrite of any old files.
    -d  Install into this dir.
EOU
}

# ---------------------------------------------------------------------



